@@ -37,6 +37,7 @@ Calamares::JobResult PackageSelectProcess::runAptCommand(const QString& command,
                                                           double endProgress,
                                                           bool verboseProgress)
 {
+    qDebug() << "Running apt command:" << command;
     QProcess aptProcess(this);
     aptProcess.setProgram("/usr/sbin/chroot");
     aptProcess.setArguments({ rootMountPoint, "/bin/bash", "-c", command });
@@ -54,6 +55,8 @@ Calamares::JobResult PackageSelectProcess::runAptCommand(const QString& command,
         commandHRPrefix = tr("Upgrading installed system: ");
     } else if (command.contains("remove")) {
         commandHRPrefix = tr("Cleaning up packages: ");
+    } else if (command.contains("cdrom")) {
+        commandHRPrefix = tr("cdrom: ");
     }
 
     QRegularExpression getRegex(R"(Get:\d+\s+[^ ]+\s+[^ ]+\s+(.+?)\s+\S+\s+(\S+)\s+\[(.*?)\])");
@@ -278,7 +281,8 @@ Calamares::JobResult PackageSelectProcess::exec()
 
     const QVariantMap installationData = gs->value("installation_data").toMap();
     const QString installationMode = installationData.value("installation_mode").toString();
-    const bool downloadUpdates = installationData.value("download_updates").toBool();
+    const bool hasInternet = installationData.value("hasInternet").toBool();
+    const bool downloadUpdates = (installationData.value("download_updates").toBool() && hasInternet);
     const QVariantList packagesToInstall = installationData.value("packages_to_install").toList();
     const QVariantList packagesToRemove = installationData.value("packages_to_remove").toList();
     const QVariantList presentSnaps = installationData.value("present_snaps").toList();
@@ -296,12 +300,35 @@ Calamares::JobResult PackageSelectProcess::exec()
     };
 
     const QVector<ProgressAllocation> allocations = allocationMap.value(installationMode, { {0.0, 1.0} });
-
-    // Run apt update
     const double aptRange = allocations[0].end - allocations[0].start;
     const double updateStart = allocations[0].start;
     const double updateEnd = updateStart + 0.1 * aptRange;
 
+    // Temporarily copy ubuntu.sources elsewhere, if we do not have network
+    // This is so we can update the apt cache safely
+    // FIXME: there has to be a better, more native way to do this. It works
+    // for now, but in the 25.10 cycle, we're probably going to move some of
+    // these command-line apt calls to the libapt C library. LP: #2107287
+    if (!hasInternet) {
+        const QString ubuntu_sources_path = rootMountPoint + "/etc/apt/sources.list.d/ubuntu.sources";
+        QFile* ubuntu_sources = new QFile(ubuntu_sources_path);
+        // Just in case this module is used in a non-Ubuntu environment, make sure ubuntu.sources exists
+        // TODO: make this configurable in the 25.10 cycle
+        if (ubuntu_sources->exists()) {
+            const QString backup_name = ubuntu_sources_path + ".bak";
+            if (!ubuntu_sources->rename(ubuntu_sources_path + ".bak")) {
+                return Calamares::JobResult::error(tr("Internal Error"),
+                                                   tr("Permission denied when moving ubuntu.sources to prepare for offline install"));
+            }
+            Calamares::JobResult addCdromResult = runAptCommand("apt-cdrom add -m -d=/media/cdrom/", rootMountPoint, updateStart, updateEnd, true);
+            if (!addCdromResult) return std::move(addCdromResult);
+        } else {
+            return Calamares::JobResult::error(tr("Internal Error"),
+                                               tr("/etc/apt/sources.list.d/ubuntu.sources not found in the target, are you a downstream?"));
+        }
+    }
+
+    // Run apt update
     m_prettyStatus = tr("Updating apt cache");
     emit prettyStatusMessageChanged(m_prettyStatus);
     emit progress(updateStart);
@@ -403,6 +430,8 @@ Calamares::JobResult PackageSelectProcess::exec()
                                                                true);
             if (!cpbe->remove()) qDebug() << "Warning: failed to clean up /usr/libexec/checkpackage-backend";
             if (!installResult) return std::move(installResult);
+        } else {
+            if (!cpbe->remove()) qDebug() << "Warning: failed to clean up /usr/libexec/checkpackage-backend";
         }
     }
     else qDebug() << "No packages to install.";
@@ -429,16 +458,14 @@ Calamares::JobResult PackageSelectProcess::exec()
     const double removeEnd = removeStart + 0.2 * aptRange;
 
     if (!removeDebPackages.isEmpty()) {
-        const QString removeCommand = QString("DEBIAN_FRONTEND=noninteractive apt-get -y --purge remove %1")
+        const QString removeCommand = QString("DEBIAN_FRONTEND=noninteractive apt-get -y --purge remove $(/usr/libexec/checkpackage-backend %1);")
                                          .arg(removeDebPackages.join(" "));
         Calamares::JobResult removeResult = runAptCommand(removeCommand,
                                                           rootMountPoint,
                                                           removeStart,
                                                           removeEnd,
                                                           true);
-        if (!removeResult) { // Using operator bool() to check for errors
-            return std::move(removeResult); // Move to avoid copy
-        }
+        if (!removeResult) return std::move(removeResult);
     }
 
     const double autoremoveStart = removeEnd;
@@ -453,6 +480,44 @@ Calamares::JobResult PackageSelectProcess::exec()
     // Disable diversions
     //divert(false);
 
+    // Move ubuntu.sources back, and clean up the cdrom file
+    // FIXME: there has to be a better, more native way to do this. It works
+    // for now, but in the 25.10 cycle, we're probably going to move some of
+    // these command-line apt calls to the libapt C library. LP: #2107287
+    try {
+    if (!hasInternet) {
+        const QString ubuntu_sources_path = rootMountPoint + "/etc/apt/sources.list.d/ubuntu.sources";
+        const QString backup_name = ubuntu_sources_path + ".bak";
+        QFile* ubuntu_sources = new QFile(ubuntu_sources_path);
+        QFile* ubuntu_sources_bak = new QFile(backup_name);
+        // Just in case this module is used in a non-Ubuntu environment, make sure ubuntu.sources exists
+        // TODO: make this configurable in the 25.10 cycle
+        if (ubuntu_sources->exists()) {
+            if (!ubuntu_sources->remove()) {
+                return Calamares::JobResult::error(tr("Internal Error"),
+                                                   tr("/etc/apt/sources.list.d/ubuntu.sources already exists and it won't budge - this is a rare edge case, please report!"));
+            }
+        }
+        if (!ubuntu_sources_bak->rename(ubuntu_sources_path)) {
+            return Calamares::JobResult::error(tr("Internal Error"),
+                                               tr("Permission denied when moving ubuntu.sources back after offline install"));
+        }
+
+        // Remove the apt-cdrom entry we added earlier
+        // This may seem drastic, but we already expect that the automirror
+        // module, ran before this, creates a deb822-style ubuntu.sources
+        QFile* cdrom_sources_list = new QFile(rootMountPoint + "/etc/apt/sources.list");
+        if (!cdrom_sources_list->remove()) {
+            return Calamares::JobResult::error(tr("Internal Error"),
+                                               tr("Failed to remove classic sources.list file"));
+        }
+    }
+    } catch (const std::exception &exc) {
+        qDebug() << exc.what();
+    } catch (...) {
+        qDebug() << "Caught unknown error";
+    }
+
     // Handle snap packages
     if (installationMode != "minimal") {
         QStringList snapPackages;
